# coding: utf-8
# flake8: noqa

"""
Matplotlib contrib functionality.
"""


__all__ = ["MatplotlibFormatter"]


# provisioning imports
from law.contrib.matplotlib.formatter import MatplotlibFormatter
