# coding: utf-8
# flake8: noqa

"""
Singularity functionality and sandbox implementation.
"""


__all__ = ["SingularitySandbox"]


# provisioning imports
from law.contrib.singularity.sandbox import SingularitySandbox
