# coding: utf-8

"""
Build large-scale task workflows using luigi, remote job submission, remote targets, and environment
sandboxing.
"""

__author__ = "Marcel Rieger"
__email__ = "github.riga@icloud.com"
__copyright__ = "Copyright 2018-2019, Marcel Rieger"
__credits__ = ["Marcel Rieger"]
__contact__ = "https://github.com/riga/law"
__license__ = "BSD-3-Clause"
__status__ = "Development"
__version__ = "0.0.33"
